from neuron import h
import matplotlib.pyplot as plt
from scipy import signal


#This class takes a NEURON HOC file as an input creates a wrapper
#which can be run by sbi for simulation data.
class Cell:

    #Constructs a Cell object.
    #Takes:
    #   1) The HOC template directory.
    #   2) The HOC object name for the given cell.
    #   3) The summary statistics function.
    def __init__(self, template_dir, template_name):
        
        #Load the template directory.
        h.load_file(template_dir)

        #Get the cell from the h object.
        invoke_cell = getattr(h, template_name)
       
        #Exctract the cell object itself.
        self.__cell = invoke_cell()

        #Initialize vectors to track membrane voltage.
        self.__mem_potential = h.Vector()
        self.__time = h.Vector()

        #Record time and membrane potential.
        self.__time.record(h._ref_t)
        self.__mem_potential.record(self.__cell.soma[0](0.5)._ref_v) #This line means that all cell templates must
                                                                     #have a soma array with at least one soma
                                                                     # element in it.

    #Return the neuron cell onject.
    def get_cell(self):
        return self.__cell

    #Return the membrane potential neuron vector as a numpy array.
    def get_potential_as_numpy(self):
        return signal.resample(self.__mem_potential.as_numpy(),32**2)
    
    #Return the time vector as a numpy array.
    def get_time_as_numpy(self):
        return signal.resample(self.__time.as_numpy(),32**2)

    #Graph the membrane vs time graph based on whatever is in membrane voltage 
    #and time vectors.
    def graph_potential(self, save_img_dir = None):
        plt.close()
        plt.figure(figsize = (20,5))
        plt.plot(self.__time, self.__mem_potential)
        plt.xlabel('Time')
        plt.ylabel('Membrane Potential')
        
        if save_img_dir != None:
            plt.savefig(save_img_dir)
        
        plt.show()


from neuron import h
import torch
from sbi import utils as utils
from sbi.inference import SNPE, prepare_for_sbi, simulate_for_sbi
import matplotlib.pyplot as plt
from sbi.utils.get_nn_models import posterior_nn
import numpy as np
torch.autograd.set_detect_anomaly(True)

#The job of the optimizer is to take some summary statistics and return a set of parameters which 
#are generated by SBI.
class Optimizer():
    
    #This function initializes SBI and sets the summary_stats function and
    #The target summary stats.
    #Arguments:
    #   1) The target cell to optimize (Cell Object).
    #   2) A list of parameters to optimize.
    #   2) A parameter range tuple which consists of the lower and upper parameter values.
    #   3) a summary_stat function which calculates the summary statistics for the optimizer.
    #   4) Two optional kwargs both used for the adaptation summary stats function:
    #       1) The spike adaptation threshold.
    #       2) The spike height threshold.
    def __init__(self, cell, parameter_list, parameter_range, summary_funct, *args, **kwargs):
        #Set some parameters.
        self.__cell = cell #The cell object.
        self.summary_funct = summary_funct

        self.__summary = summary_funct != None

        #A dictionary which stores the summary stat function aditional parameters if they are 
        #needed.
        self.summary_stat_args = ()
        self.summary_stat_kwargs = kwargs['kwargs'] if 'kwargs' in kwargs else kwargs

        #Set the default simulation parameters.
        self.set_simulation_params()
        
        self.__i_clamp = h.IClamp(cell.get_cell().soma[0](0.5))

        #Set the parameter list.
        self.__cell_optimization_params = parameter_list

        #Set the parameter range for the above parameters.
        if parameter_range != None:
            lows = torch.tensor(parameter_range[0], dtype=float)
            highs = torch.tensor(parameter_range[1], dtype=float)
            self.__prior = utils.BoxUniform(low=lows, high=highs)
        else:
            self.__prior = None

        self.__posterior = []

    #This sets the simmulation parameters.
    def set_simulation_params(self, sim_run_time = 600, delay = 50, inj_time = 500, i_inj = 0.2, v_init = -75):
        self.__sim_run_time = sim_run_time
        self.__delay = delay
        self.__inj_time = inj_time
        self.__i_inj = i_inj
        self.__v_init = v_init
        self.__steps_per_ms = 5000 / sim_run_time #NOTE: The 96**2 has to be fixed because of the CNN implemenation.
                                                       #This fixes the number of data points to be 32^2.
        

    def set_simulation_optimization_params(self, param_list):
        self.__cell_optimization_params = param_list

    def set_current_injection_list(self, current_injections):
        self.__current_injections = current_injections

    def get_simulation_optimization_params(self):
        return self.__cell_optimization_params

    #Returns a tuple countaining simulation time variables.
    def get_simulation_time_varibles(self):
        return (self.__sim_run_time, self.__delay, self.__inj_time)

    def set_target_statistics(self, stats):
        self.__observed_stats = stats

    def graph_performance(self, posterior_id, sample_threshold=1000):
        samples = self.__posterior[posterior_id].sample((sample_threshold,), x=self.__observed_stats)
        fig, axes = utils.pairplot(samples,
                        fig_size=(5,5),
                        points_offdiag={'markersize': 6},
                        labels=self.__cell_optimization_params,
                        points_colors='r');
        plt.tight_layout()
        plt.show()

    def get_best_sample(self, posterior_id): 
        return self.__posterior[posterior_id].sample((1,), x=self.__observed_stats).numpy()[0]

    def get_samples(self, posterior_id, sample_threshold):
        return self.__posterior[posterior_id].sample((sample_threshold,), x=self.__observed_stats).numpy()

    #This is the function which is called by SBI to actually generate the sample distribution.
    #This function takes in some args and kwargs depending on what the function is being
    #used for.
    #If there are any args:
    #   1) The first argument must be a list of the parameters
    #      to be set for this simulation. These parameters must appear 
    #      in order based on the internaly set cell_opti

    def get_samples(self, posterior_id, sample_threshold):
        return self.__posterior[posterior_id].sample((sample_threshold,), x=self.__observed_stats).numpy()

    #This is the function which is called by SBI to actually generate the sample distribution.
    #This function takes in some args and kwargs depending on what the function is being set for all sections 
    #      in the cell.
    #
    #      The args list is mainly used by SBI to run the simulations.
    #      When simulation wrapper is called by SBI a list of parameters is passed
    #      in corresponding to the previously specified parameters list.
    #
    #If there are any kwargs:
    #   - Each kwarg corresponds to a parameter to be set. For instance,
    #   kwargs['gbar_natCA3'] = 0.1 would mean set cell.gbar_natCA3 = 0.1.
    def simulation_wrapper(self, *args, **kwargs):
        #Set simulation parameters.
        h.tstop = self.__sim_run_time
        h.v_init = self.__v_init
        h.dt = 1/self.__steps_per_ms
        h.steps_per_ms = self.__steps_per_ms
        
        #Set current clamp values.
        self.__i_clamp.dur = self.__inj_time
        self.__i_clamp.amp = self.__i_inj
        self.__i_clamp.delay = self.__delay
        

        #Set parameters based on the parameters list.
        cell = self.__cell.get_cell()
        if self.__cell_optimization_params != None:
            for sec in cell.all:
                for index, key in enumerate(self.__cell_optimization_params):
                    setattr(sec, key, args[0][index]) 

        #Set cell parameters in all sections based on the kwargs.
        for sec in cell.all:
            for key in kwargs:
                setattr(sec, key, kwargs[key]) 

        #Run the simulation with the given parameters.
        h.run()

        #Pass in the summary stat args and kwargs if they exist.
        #Take care of all possible cases.
        #If the summary stat funct is user defined, then give the user the whole cell for their
        #summary stats function. Otherwise, pass in a tensor of just the voltage trace data
        # (for the CNN).
       
        
        if self.__summary:
            voltage = self.__cell.get_potential_as_numpy()
            time    = self.__cell.get_time_as_numpy() 
            return self.summary_funct(voltage, time, *self.summary_stat_args, **self.summary_stat_kwargs)
        
        data = torch.from_numpy(self.__cell.get_potential_as_numpy()).float()
        return data

    def multi_channel_wrapper_summary(self, *args, **kwargs):
        data = np.array([])

        for current_injection in self.__current_injections:
            self.set_simulation_params(i_inj=current_injection)
            data = np.concatenate((data,self.simulation_wrapper(*args, **kwargs)), axis=None)
        
        return data

    def multi_channel_wrapper_CNN(self, *args, **kwargs):
        data = torch.empty((len(self.__current_injections),1024))
        
        for index, current_injection in enumerate(self.__current_injections):
            self.set_simulation_params(i_inj=current_injection)
            data[index] = self.simulation_wrapper(*args, **kwargs)
        
        return data

    def set_observed_stats(self, stats):
        self.__observed_stats = stats

    #This function builds simulation data ONLINE.
    #This is what uses SBI to infer the parameters with the above simulation wrapper.
    #This function takes three arguments:
    #   1) The number of simulations to run per round.
    #   2) The number of workers to use to run these simulations. NOTE: This is currently broken and 
    #      anything except workers=1 will cause the program to crash. Refer to this issue for more
    #      details as to why this happens: https://www.mackelab.org/sbi/faq/question_03/
    #   3) The number of rounds for inference. Each round a posterior distribution is generated
    #      which is used as the prior for the next round, hopefully converging on an even better
    #      distribution than just one big round.
    #   4) The CNN which is used to learn the summary stats, if this is set to None no embedding net
    #      exists and this step can be skipped.
    def run_inference_multiround(self, num_simulations=1000, num_rounds = 1, workers=1):
        #Get stuff ready for sbi.
        simulator, self.__prior = prepare_for_sbi(self.multi_channel_wrapper_summary, self.__prior)
        self.__inference = SNPE(prior=self.__prior)
        
        proposal = self.__prior

        for _ in range(num_rounds):
            theta, x = simulate_for_sbi(simulator, proposal, num_simulations=num_simulations,num_workers=workers)
            density_estimator = self.__inference.append_simulations(theta, x, proposal=proposal)
            density_estimator.train(show_train_summary=True)

            proposal = self.__inference.build_posterior()
            self.__posterior.append(proposal)

            #Set defualt x.
            #Take maximum probability here. Potentially.
            samples = self.__posterior[-1].sample((1000,), x=self.__observed_stats)
            log_prob = self.__posterior[-1].log_prob(samples, x=self.__observed_stats, norm_posterior=False)
            proposal = samples[np.argmax(log_prob)]
        
        
    def run_inference_learned_stats(self, embedding_net, num_simulations=1000, num_rounds=1, workers=1):        
        
        #Get stuff ready for sbi.
        self.__simulator, self.__prior = prepare_for_sbi(self.multi_channel_wrapper_CNN, self.__prior)
        
        neural_posterior = utils.posterior_nn(model='maf', 
                                                embedding_net=embedding_net,
                                                hidden_features=10,
                                                num_transforms=2)
        self.__inference = SNPE(prior=self.__prior, density_estimator=neural_posterior)

        proposal = self.__prior

        for _ in range(num_rounds):
            #Do the first round so we train the weights for the CNN.
            theta, x = simulate_for_sbi(self.__simulator, self.__prior, num_simulations=num_simulations,num_workers=workers)
            density_estimator = self.__inference.append_simulations(theta, x)
            density_estimator.train(show_train_summary=True)
            proposal = self.__inference.build_posterior()
            self.__posterior.append(proposal)

            #Set defualt x.
            #Take maximum probability here. Potentially.
            samples = self.__posterior[-1].sample((1000,), x=self.__observed_stats)
            log_prob = self.__posterior[-1].log_prob(samples, x=self.__observed_stats, norm_posterior=False)
            proposal = samples[np.argmax(log_prob)]

    def clear_posterior(self):
        self.__posterior = []


import numpy as np
from asct.src.Tuner import CellTuner
import argparse
import json
import csv
from datetime import datetime
import importlib
import os

#Takes a string and appends it to a log file.
#This is terribly inefficient but it shouldnt really matter because we arent writing that much info.
def log_ouput(s_input, file_dir):
    file = open(file_dir, 'a')
    file.write(s_input)
    file.close()

def tune_with_template(config_data, *args, **kwargs):


    manifest = config_data['manifest']
    conditions = config_data['conditions']
    run = config_data['run']
    settings = config_data['optimization_settings']
    parameters = config_data['optimization_parameters']
    
    summary = {}
    summary_function = None

    log = kwargs['log']
    log_name = 'tuning_logs/tuning_log_%s.log' % datetime.now().strftime("%m_%d_%Y|%H:%M:%S")
    
    if log:
        if not os.path.isdir('tuning_logs'):
            os.mkdir('tuning_logs')

    if 'summary' not in config_data and manifest['architecture'] == 'summary':
        print('Invalid config file.')
        return

    elif 'summary' in config_data:
        summary = config_data['summary']

        #Lets load the summary function into memory.
        summary['summary_file'] = summary['summary_file'].replace('.py','').replace('/','.')

        summary_module = importlib.import_module(summary['summary_file'])
        summary_function = getattr(summary_module, summary['function_name'])

        #Now remove the file and function name from the dictionary.
        del summary['summary_file']
        del summary['function_name']

    if not kwargs['c_mod']:
        modfiles_dir = None
    
    tuner = CellTuner( 
                      manifest['modfiles_dir'], 
                      manifest['template_dir'], 
                      manifest['template_name'],
                      parameters['current_injections'],
                      parameters['parameters'], 
                      (parameters['lows'], parameters['highs']), 
                      architecture=manifest['architecture'],
                      summary_funct=summary_function,
                      features=settings['features'],
                      kwargs=summary)       #Pass in the summary function variables as kwargs

    tuner.set_simulation_params(sim_run_time=run['tstop'], delay=run['delay'],inj_time=run['duration'],v_init=conditions['v_init'])
    
    #We need to check if we are validating against a ground truth model or if we are tuning from input data.
    if manifest['job_type'] == 'ground_truth':
        tuner.calculate_target_stats_from_model(manifest['target_template_dir'], manifest['target_template_name'])
        
        tuner.optimize_current_injections(num_simulations=settings['num_simulations'],
                                          inference_workers=settings['workers'], 
                                          sample_threshold=kwargs['result_threshold'], 
                                          num_rounds=settings['num_rounds'])

        tuner.find_best_parameter_sets()
        
        s_out = 'The optimizer found the following parameter set:\n' + str(tuner.get_optimial_parameter_sets(kwargs['result_threshold']))
        print(s_out)
        if log:
            log_ouput(s_out,log_name)

        s_out = 'Relative Percent Error from ground truth:\n' + str(tuner.compare_found_parameters_to_ground_truth(parameters['ground_truth'],kwargs['result_threshold']))
        print(s_out)
        if log:
            log_ouput(s_out,log_name)

        tuner.generate_target_from_model()
        tuner.compare_found_solution_to_target(kwargs['result_threshold'],kwargs['display'],kwargs['save_dir'])
    
    elif manifest['job_type'] == 'from_data':
        #Load in the input_data
        responses = load_current_injections_from_csv(manifest['input_data'])
        tuner.set_target_responses(responses)

        tuner.calculate_target_stats_from_data()
        
        tuner.optimize_current_injections(num_simulations=settings['num_simulations'],
                                          inference_workers=settings['workers'], 
                                          sample_threshold=kwargs['result_threshold'], 
                                          num_rounds=settings['num_rounds'])

        tuner.find_best_parameter_sets()
        s_out = 'The optimizer found the following parameter set:\n' + str(tuner.get_optimial_parameter_sets(kwargs['result_threshold']))
        print(s_out)
        if log:
            log_ouput(s_out,log_name)

        tuner.compare_found_solution_to_target(kwargs['result_threshold'],kwargs['display'],kwargs['save_dir'])

def parse_config(config_directory):
    data = None

    try:
        file = open(config_directory)
    except FileNotFoundError:
        print('The config file %s could not be found.' % config_directory)
    else:
        data = json.load(file)

    return data

def load_current_injections_from_csv(file_dir):
    file = open(file_dir, 'r')
    csv_file =  csv.reader(file,delimiter=',')

    responses = [[] for _ in range(len(next(csv_file)))]
    file.seek(0)
    for row in csv_file:
        for i, column in enumerate(row):
            responses[i].append(column)
    
    responses = [np.array(e,dtype=float) for e in responses]
    return responses

def main():
    argument_parser = argparse.ArgumentParser(description='Uses SBI to find optimal parameter sets for biologically realistic neuron simulations.')

    argument_parser.add_argument('config_dir', type=str, help='the optimization config file directory')
    argument_parser.add_argument('save_dir', nargs='?', type=str, default=None, help='[optional] the directory to save figures to')
    argument_parser.add_argument('-g', default=False, action='store_true', help='displays graphics')
    argument_parser.add_argument('-c', default=False, action='store_true', help='compiles modfiles')
    argument_parser.add_argument('-l', default=False, action='store_true', help='store log files')
    argument_parser.add_argument('-n', default=1, type=int, help='the number of found parameters to show (must be less than the threshold sample size in the optimization config file)')

    args = argument_parser.parse_args()

    config_data = parse_config(args.config_dir)

    tune_with_template(config_data, save_dir=args.save_dir, display=args.g, result_threshold=args.n, c_mod=args.c, log=args.l)

if __name__ == '__main__':
    main()
    


from asct.src.Optimizer import Optimizer
from asct.src.Cell import Cell
from asct.src.SummaryNet import SummaryCNN
from neuron import h
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import math
import torch
from tqdm import tqdm


#This class uses the Cell and Optimizer classes to
#optimize a set of current injections based on
#either some user set target statistics or a 
#target model.
#Takes 7 parameters:
# 1) A list of the current injection levels (nA) to match.
# 2) The directory of the modfiles for used.
# 3) The directory for the hoc template file.
# 4) The name of the target template in the template file.
# 5) A list of each parameter to optimize. These should be names of parameters which can be accessed in 
#    the hoc object.
# 6) A tuple containing two arrays, the lows and highs for each parameter specified in the above argument.
# 7) A bool which tells the optimizer to use a CNN to learn the summary statistics or not. If not set then 
#    the tuner uses a user defined summary stats function.
# 8) Some optional kwargs which contain values used for the summary stats function if the parameters
#    arent being learned and aditional parameters need to be passed to the summary stats function.


class CellTuner:
    def __init__(self, modfiles_dir, template_dir, template_name, current_injections, optimization_parameters, parameter_range, architecture='summary', summary_funct = None, features=8,*args, **kwargs):
        
        #Store the current injection list and the spike threshold.
        self.__current_injections = current_injections

  
        self.__embedding_net = None
        if architecture == 'convolution' or architecture == 'hybrid':
            self.__embedding_net = SummaryCNN(len(current_injections), features, hybrid= False if architecture == 'convolution' else True)
            summary_funct = None

            args = ()
            kwargs = {}
        

        if modfiles_dir != None:
            #First lets try to compile our modfiles.
            if os.system('nrnivmodl %s' % modfiles_dir) == 0:
                print("Compiled Successfully.")
            else:
                print("Could not compile the modfiles at %s" % modfiles_dir)

        #Now load in the standard run hoc.
        h.load_file('stdrun.hoc')

        #Set up the cell and the optimizer. This will be responsible for optimizing the given cell
        #at each current injection value.
        self.__to_optimize = Cell(template_dir, template_name)

        self.__optimizer = Optimizer(self.__to_optimize, optimization_parameters, parameter_range, summary_funct, *args, **kwargs)

        self.__parameter_samples = []
        self.__target_responses = None

        self.NUM_SAMPLES = 1000

    def run_forward_pass(self, input):
        out = self.__embedding_net(input)
        return out
    
    #Set the simulation parameters.
    def set_simulation_params(self, sim_run_time = 1500, delay = 400, inj_time = 500, v_init = -75):
        self.__sim_run_time = sim_run_time
        self.__delay = delay
        self.__inj_time = inj_time
        self.__v_init = v_init

    #Calculate the target summary statistics based on a provided HOC model.
    #If CNN is activated and not yet trained, train it.
    def calculate_target_stats_from_model(self, template_dir, template_name):    
        self.__target_cell = Cell(template_dir, template_name)
        
        #Create a temporary optimizer which is used to get the target
        #statistics from the target cell at each current injection.
        self.__sim_environ = Optimizer(self.__target_cell, 
                                      None, 
                                      None,
                                      self.__optimizer.summary_funct,
                                      *self.__optimizer.summary_stat_args,
                                      **self.__optimizer.summary_stat_kwargs)

        self.__sim_environ.set_current_injection_list(self.__current_injections)
        
        if self.__embedding_net == None:
            self.__optimizer.set_observed_stats(self.__sim_environ.multi_channel_wrapper_summary())
        else:
            self.__optimizer.set_observed_stats(self.__sim_environ.multi_channel_wrapper_CNN().flatten())

    #Given the current injection responses directly, save it.
    def set_target_responses(self, target_responses):
        self.__time = np.linspace(0,self.__sim_run_time * 1e-3,1024)
        self.__target_responses = target_responses

    #Generates the target statitcs from the data. Target responses must be set.
    def calculate_target_stats_from_data(self):
        if self.__embedding_net == None:
            
            #We just need to get the statistics for each current injection.
            stats = np.array([])

            for trace in self.__target_responses:
                stat = self.__optimizer.summary_funct(trace, self.__time, *self.__optimizer.summary_stat_args, **self.__optimizer.summary_stat_kwargs)
                stats = np.concatenate((stats, stat), axis=0)

            self.__optimizer.set_observed_stats(stats)
        else:
            data = torch.empty((len(self.__current_injections),1024))
            for index,response in enumerate(self.__target_responses):
                data[index] = torch.from_numpy(response).float()

            self.__optimizer.set_observed_stats(data.flatten())

    #Wrapper function which automatically chooses which function to call based on
    #set parameters.
    def optimize_current_injections(self, num_simulations = 500, num_rounds=1, inference_workers=1 ,sample_threshold = 10):
        self.__top_n = sample_threshold
        
        if self.__embedding_net == None:
            self.optimize_current_injections_summary(num_simulations=num_simulations, 
                                                     num_rounds=num_rounds, 
                                                     inference_workers=inference_workers)
        else:
            self.optimize_current_injections_cnn(num_simulations=num_simulations, 
                                                 inference_workers=inference_workers,
                                                 num_rounds=num_rounds)
    
    def optimize_current_injections_cnn(self, num_simulations = 500, num_rounds = 1, inference_workers=1):        
        self.__optimizer.set_current_injection_list(self.__current_injections)
        self.__optimizer.run_inference_learned_stats(self.__embedding_net, num_simulations=num_simulations, num_rounds=num_rounds, workers=inference_workers)
        self.__parameter_samples = self.__optimizer.get_samples(-1, sample_threshold=self.NUM_SAMPLES)

    def optimize_current_injections_summary(self, num_simulations = 500, num_rounds=1, inference_workers=1):
        self.__optimizer.set_current_injection_list(self.__current_injections)
        self.__optimizer.run_inference_multiround(num_simulations=num_simulations, workers=inference_workers, num_rounds=num_rounds)
        self.__parameter_samples = self.__optimizer.get_samples(-1, sample_threshold=self.NUM_SAMPLES)
            

    def compute_correlation_for_parameter_set(self, parameter_set):
        #generate the found voltage responses.
        found_responses = []
        for i_inj in self.__current_injections:
            self.__optimizer.set_simulation_params(sim_run_time=self.__sim_run_time, delay=self.__delay, inj_time=self.__inj_time, v_init=self.__v_init, i_inj=i_inj)
            self.__optimizer.simulation_wrapper(parameter_set)
            found_responses.append(np.copy(self.__to_optimize.get_potential_as_numpy()))
        
       

        #Compute average cosine similarity accross current injections.
        num_injections = len(self.__current_injections)
        mean_similarity = 0
        for i in range(num_injections):
            mean_similarity += np.dot(found_responses[i], self.__target_responses[i]) / (np.linalg.norm(found_responses[i]) * np.linalg.norm(self.__target_responses[i]))
        mean_similarity /= num_injections
        
        return mean_similarity


    #Find the best parameter set based on the calulated posterior distributions.
    #This is done by getting the top 'x' solutions from each posterior distribution
    #and finding the closest match in each set of parameters.
    #Takes one parameter:
    #   1) Displays the best found set and its matching ratio.
    #
    #NOTE: as of now there is nothing which takes into account the rank of the parameters,
    #i.e. all parameter sets from out of the top 'x' are treated equally when in reality the
    #first ones are really the best solutions, this could be taken into account in a better
    #function
    #Find the "top_n best paramter sets from the posterior"
    def find_best_parameter_sets(self, SHOW_TOP_CORRELATION=True):
                 
        #Make sure the target is generated if it has not been set.
        if self.__target_responses == [] or self.__target_responses == None:
            self.generate_target_from_model()

        #sort all the paramter samples based on performance (cosine correlation.)
        parameter_ranking = []
        for pset in tqdm(self.__parameter_samples,desc='Generating sample performance for %d samples' % self.NUM_SAMPLES):
            parameter_ranking.append((self.compute_correlation_for_parameter_set(pset), pset))
        
        #Now sort the parameter sets.
        parameter_ranking.sort(key = lambda x:x[0], reverse=True)

        #Now get the top_n samples.
        self.__found_parameters_correlation = [elem[0] for elem in parameter_ranking[:self.__top_n]]
        self.__found_parameters = [elem[1] for elem in parameter_ranking[:self.__top_n]]
        
        if SHOW_TOP_CORRELATION:
            print('Top sample correlation performance: ', self.__found_parameters_correlation)

    #Compare the found solution to the model using the target cell.
    #Takes 2 parameters:
    #   1) If a graph should be displayed which shows the voltage trace of
    #      found parameter set overlayed on the target cell.
    #   2) The directory the above image should be saved to. None means dont
    #      save the image.
    def compare_found_solution_to_target(self, top_n=1, display=False, save_dir=None):
        #Get the time vector.
        #create a gid of subplots to display the results. the grid is NxN where
        #n = ceil(sqrt(n))
        n_plots = math.ceil(math.sqrt(top_n))
        fig = plt.figure(figsize=(10,10))
        outer = gridspec.GridSpec(n_plots,n_plots)

        for i in range(top_n):
            #Now get the real cell with our current injections.
            found_responses = []
            for i_inj in self.__current_injections:
                self.__optimizer.set_simulation_params(sim_run_time=self.__sim_run_time, delay=self.__delay, inj_time=self.__inj_time, v_init=self.__v_init, i_inj=i_inj)
                self.__optimizer.simulation_wrapper(self.__found_parameters[i])
                found_responses.append(np.copy(self.__to_optimize.get_potential_as_numpy()))
            
            #Now plot everything.
            current_injection_length = len(self.__current_injections)
            inner = gridspec.GridSpecFromSubplotSpec(current_injection_length, 1, subplot_spec=outer[i])
            if current_injection_length > 1:
                for index in range(current_injection_length):
                    ax = plt.Subplot(fig,inner[index])
                    if index == 0:
                        ax.set_title("Solution: %d" % (i+1))
                    if index != current_injection_length - 1:
                        ax.xaxis.set_visible(False)
                    ax.plot(self.__time, self.__target_responses[index], label='Target')
                    ax.plot(self.__time, found_responses[index], label='Found')
                    ax.set_ylabel(self.__current_injections[index])
                    fig.add_subplot(ax)
            else:
                ax = plt.Subplot(fig,inner[0])
                ax.plot(self.__time, self.__target_responses[0], label='Target')
                ax.plot(self.__time, found_responses[0], label='Found')
                ax.set_ylabel(self.__current_injections[0])
                fig.add_subplot(ax)

        handles, labels = plt.gca().get_legend_handles_labels()
        fig.legend(handles, labels, loc='upper right')
        fig.tight_layout()

        if display:  
            fig.show()
        
        if save_dir != None:
            fig.savefig(save_dir,bbox_inches='tight')

    def generate_target_from_model(self):
        #Find the target voltage traces.
        self.__target_responses = []
        for i_inj in self.__current_injections:
            self.__sim_environ.set_simulation_params(sim_run_time=self.__sim_run_time, delay=self.__delay, inj_time=self.__inj_time, v_init=self.__v_init, i_inj=i_inj)
            self.__sim_environ.simulation_wrapper()
            self.__target_responses.append(np.copy(self.__target_cell.get_potential_as_numpy()))

        #Store this for other functions which need the time vector.
        self.__time = self.__to_optimize.get_time_as_numpy()

    #def generate_found_FI_curve(self, display=False, save_dir=None):

    #Return a string containing the parameter names as keys and the found parmeter as the values
    #for the top n values.
    def get_optimial_parameter_sets(self, top_n=1):        
        p_list = []
        for index in range(top_n):
            p_list.append(dict(zip(self.__optimizer.get_simulation_optimization_params(), self.__found_parameters[index])))

        formated_output = ''
        
        for index, solution in enumerate(p_list):
            formated_output = formated_output + 'Solution %d: ' % (index + 1) + str(solution) + '\n'
        
        return formated_output

    #Compare the found parameter sets to the ground truth parameters.
    def compare_found_parameters_to_ground_truth(self, ground_truth, top_n=1):
        found_parameters = self.get_optimial_parameter_sets(top_n=top_n)
        
        error_list = []

        for i in range(len(found_parameters)):
            rpe = {}
            for index, parameter_name in enumerate(found_parameters[i]):
                rpe[parameter_name] = ((found_parameters[i][parameter_name] - ground_truth[index]) / ground_truth[index]) * 100
            
            average_rpe = sum([abs(rpe[key]) for key in rpe])/ len(rpe)  #Calculate relative percent error.

            error_list.append((average_rpe, rpe))

        return error_list
    
    #Compare the best found solution to the target voltage trace and return the error.
    #This returns the average cosine similiarity between all found and target voltage traces.
    def get_best_trace_error(self):
        return 1.0 - self.compute_correlation_for_parameter_set(self.__found_parameters[0])




from setuptools import setup
from setuptools import find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="asct",
    version="0.0.1",
    author="Pete Canfield",
    author_email="pbczgf@umsystem.edu",
    description="Automated Single Cell Tuner",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pbcanfield/ASCT",
    download_url='',
    license='MIT',
    install_requires=[
        'sbi',
        'neuron',
        'numpy',
        'scipy',
        'torch',
        'tqdm'
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        "Operating System :: OS Independent",
    ],
    packages=find_packages(exclude=['tests']),
    entry_points={
        'console_scripts': [
            'asct = asct.optimize_cell:main'
        ]
    }
)


import torch.nn as nn 
import torch.nn.functional as F 
import torch
import numpy as np
from scipy import stats


#Information to calculate 1D CNN output size and maxpool output size.
#https://towardsdatascience.com/pytorch-basics-how-to-train-your-neural-net-intro-to-cnn-26a14c2ea29
#Maxpool layer uses same resource but kernel size and stride are set to be the same
#unless otherwise specified. Padding set to 0 by default.
class SummaryCNN(nn.Module):
    def __init__(self, current_injections=1, summary_features=8, hybrid=True): 
        super().__init__()

        self.__hybrid = hybrid
        self.__current_injections = current_injections

        # 1D convolutional layer
        #1024->1024 
        self.conv1 = nn.Conv1d(in_channels=current_injections, out_channels=5, kernel_size=9, padding=4,stride=1)
        # Maxpool layer that reduces 1024-> 512
        self.pool1 = nn.MaxPool1d(kernel_size=2, stride=2)

        #Second convolutional layer.
        #512 -> 512
        self.conv2 = nn.Conv1d(in_channels=5, out_channels=5, kernel_size=7, padding=3, stride=1)
        #Maxpool layer to reduce from 512 to 256. (Maybe try 128 here).
        self.pool2 = nn.MaxPool1d(kernel_size=2, stride=2)

        #Third convolutional layer.
        #256 -> 256
        self.conv3 = nn.Conv1d(in_channels=5, out_channels=5, kernel_size=5, padding=2, stride=1)
        #Maxpool layer to reduce from 256 to 128. (Maybe try 128 here).
        self.pool3 = nn.MaxPool1d(kernel_size=2, stride=2)

        #Fourth convolutional layer.
        #128 -> 64
        self.conv4 = nn.Conv1d(in_channels=5, out_channels=1, kernel_size=3, padding=1, stride=1)
        #Maxpool layer to reduce from 256 to 128. (Maybe try 128 here).
        self.pool4 = nn.MaxPool1d(kernel_size=2, stride=2)

        # Fully connected layer taking as input the 6 flattened output arrays from the maxpooling layer
        self.fc = nn.Linear(in_features=64, out_features=summary_features) 

    def forward(self, x):
        #Reshape input to the right size. Inlcluding:
        #1) batch size,
        #2) Number of input channels.
        #3) Size of input.
        # -1 means to retain the first dimensions size (batch size).
        x = x.view(-1, self.__current_injections, 1024)

        raw_data = x

        x = self.pool1(F.relu(self.conv1(x)))
        res = self.pool2(x)
        x = self.pool2(F.relu(self.conv2(x)))
        x = x + res
        res = self.pool3(x)
        x = self.pool3(F.relu(self.conv3(x)))
        x = x + res
        res = self.pool4(x)
        x = self.pool4(F.relu(self.conv4(x)))

        x = x.view(-1, 64)
        x = self.fc(x)

        #add mean, standard deviation, skew, 4th moment maybe
        #Caclulate each stat for each channel and concat into a vector
        #Then concat with final x.

        if self.__hybrid:
            x = torch.cat((x,self.calculate_hybrid_stats(raw_data)), 1)

        return x

    def calculate_hybrid_stats(self, x):
        data = x.numpy()
        mean     = np.mean(data,-1)
        variance = np.std(data,-1)
        skew     = stats.skew(data,-1)
       
        return torch.from_numpy(np.concatenate((mean,variance,skew), -1))
            